# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import json

from os import getenv
from ovos_utils.messagebus import FakeBus

from neon_minerva.skill import get_skill_object, load_resource_tests


class TestSkillResources(unittest.TestCase):
    # Static parameters
    messages = list()
    bus = FakeBus()
    bus.run_forever()
    test_skill_id = 'test_skill.test'

    # Define skill and resource spec to use in tests
    resources = load_resource_tests(getenv("RESOURCE_TEST_FILE"))
    skill_entrypoint = getenv("TEST_SKILL_ENTRYPOINT")

    # Specify valid languages to test
    supported_languages = resources['languages']

    # Specify skill intents as sets
    adapt_intents = set(resources['intents']['adapt'])
    padatious_intents = set(resources['intents']['padatious'])

    # regex entities, not necessarily filenames
    regex = set(resources['regex'])
    # vocab is lowercase .voc file basenames
    vocab = set(resources['vocab'])
    # dialog is .dialog file basenames (case-sensitive)
    dialog = set(resources['dialog'])

    core_config_patch = {"secondary_langs": supported_languages}

    @classmethod
    def setUpClass(cls) -> None:
        cls.bus.on("message", cls._on_message)

        cls.skill = get_skill_object(skill_entrypoint=cls.skill_entrypoint,
                                     bus=cls.bus, skill_id=cls.test_skill_id,
                                     config_patch=cls.core_config_patch)

        cls.adapt_intents = {f'{cls.test_skill_id}:{intent}'
                             for intent in cls.adapt_intents}
        cls.padatious_intents = {f'{cls.test_skill_id}:{intent}'
                                 for intent in cls.padatious_intents}

    @classmethod
    def _on_message(cls, message):
        cls.messages.append(json.loads(message))

    def test_skill_setup(self):
        self.assertEqual(self.skill.skill_id, self.test_skill_id)
        if hasattr(self.skill, "_core_lang"):
            # ovos-workshop < 0.0.15
            self.assertEqual(set([self.skill._core_lang] +
                                 self.skill._secondary_langs),
                             set(self.supported_languages),
                             f"expected={self.supported_languages}")
        else:
            self.assertEqual(set([self.skill.core_lang] +
                                 self.skill.secondary_langs),
                             set(self.supported_languages),
                             f"expected={self.supported_languages}")

    def test_intent_registration(self):
        registered_adapt = list()
        registered_padatious = dict()
        registered_vocab = dict()
        registered_regex = dict()
        for msg in self.messages:
            if msg["type"] == "register_intent":
                registered_adapt.append(msg["data"]["name"])
            elif msg["type"] == "padatious:register_intent":
                lang = msg["data"]["lang"]
                registered_padatious.setdefault(lang, list())
                registered_padatious[lang].append(msg["data"]["name"])
            elif msg["type"] == "register_vocab":
                lang = msg["data"]["lang"]
                if msg['data'].get('regex'):
                    registered_regex.setdefault(lang, dict())
                    regex = msg["data"]["regex"].split(
                        '<', 1)[1].split('>', 1)[0].replace(
                        self.test_skill_id.replace('.', '_'), '')
                    registered_regex[lang].setdefault(regex, list())
                    registered_regex[lang][regex].append(msg["data"]["regex"])
                else:
                    registered_vocab.setdefault(lang, dict())
                    voc_filename = msg["data"]["entity_type"].replace(
                        self.test_skill_id.replace('.', '_'), '').lower()
                    registered_vocab[lang].setdefault(voc_filename, list())
                    registered_vocab[lang][voc_filename].append(
                        msg["data"]["entity_value"])
        self.assertEqual(set(registered_adapt), self.adapt_intents,
                         registered_adapt)
        for lang in self.supported_languages:
            if self.padatious_intents:
                self.assertEqual(set(registered_padatious[lang]),
                                 self.padatious_intents,
                                 registered_padatious[lang])
            if self.vocab:
                self.assertEqual(set(registered_vocab[lang].keys()),
                                 self.vocab, registered_vocab)
            if self.regex:
                self.assertEqual(set(registered_regex[lang].keys()),
                                 self.regex, registered_regex)
            for voc in self.vocab:
                # Ensure every vocab file has at least one entry
                self.assertGreater(len(registered_vocab[lang][voc]), 0)
            for rx in self.regex:
                # Ensure every rx file has exactly one entry
                self.assertTrue(all((rx in line for line in
                                     registered_regex[lang][rx])), self.regex)

    def test_dialog_files(self):
        for lang in self.supported_languages:
            dialogs = self.skill._lang_resources[lang].dialog_renderer.templates
            for dialog in self.dialog:
                self.assertIn(dialog, dialogs.keys(),
                              f"lang={lang}")

    # TODO: Consider adding tests for resource file existence



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ovos_utils.log import LOG
from ovos_utils.messagebus import FakeBus

from neon_minerva.exceptions import IntentNotMatched, ConfidenceTooLow
from neon_minerva.intent_services import IntentMatch


class PadatiousContainer:
    def __init__(self, lang: str, cache_path: str, bus: FakeBus):
        from padatious import IntentContainer
        self.cache_dir = cache_path
        self.lang = lang.lower()
        self.bus = bus
        self.padatious = IntentContainer(cache_path)
        self.bus.on('padatious:register_intent', self.register_intent)
        self.bus.on('padatious:register_entity', self.register_entity)

    def register_intent(self, message):
        """Messagebus handler for registering intents.

        Args:
            message (Message): message triggering action
        """
        lang = message.data.get('lang', self.lang)
        lang = lang.lower()
        if lang == self.lang:
            LOG.debug(f"Loading intent: {message.data['name']}")
            self.padatious.load_intent(message.data['name'],
                                       message.data['file_name'])
        else:
            LOG.debug(f"Ignoring {message.data['name']}")

    def register_entity(self, message):
        """Messagebus handler for registering entities.

        Args:
            message (Message): message triggering action
        """
        lang = message.data.get('lang', self.lang)
        lang = lang.lower()
        if lang == self.lang:
            self.padatious.load_entity(message.data['name'],
                                       message.data['file_name'])

    def calc_intent(self, utt: str) -> dict:
        intent = self.padatious.calc_intent(utt)
        LOG.debug(intent)
        return intent.__dict__ if intent else dict()


class TestPadatiousMatcher:
    def __init__(self, container: PadatiousContainer,
                 include_med: bool = True, include_low: bool = False):
        LOG.debug("Creating test Padatious Matcher")
        if include_low:
            self.min_conf = 0.5
        elif include_med:
            self.min_conf = 0.8
        else:
            self.min_conf = 0.95
        self.padatious = container

    def test_intent(self, utterance: str) -> IntentMatch:
        intent = self.padatious.calc_intent(utterance)
        if not intent:
            raise IntentNotMatched(utterance)
        conf = intent.get("conf") or 0.0
        if conf < self.min_conf:
            raise ConfidenceTooLow(f"{conf} less than minimum {self.min_conf}: "
                                   f"{utterance}. intent={intent}")
        skill_id = intent.get('name').split(':')[0]
        return IntentMatch('Padatious', intent.get('name'),
                           intent.get('matches') or intent.get('entities'),
                           skill_id, utterance)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2023 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2023: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending


import unittest

from os import getenv
from time import sleep
from chatbot_core.utils import clean_up_bot
from klat_connector import start_socket
from klat_connector.mach_server import MachKlatServer
from neon_minerva.chatbots.util import load_chatbot


class TestSubmind(unittest.TestCase):
    # Initialize a server for testing
    server = MachKlatServer()
    sleep(1)
    socket = start_socket("0.0.0.0")
    submind = None

    @classmethod
    def setUpClass(cls) -> None:
        # Determine submind to test
        submind_entrypoint = getenv("TEST_BOT_ENTRYPOINT")
        bot_class = load_chatbot(submind_entrypoint)
        # Initialize a submind instance
        cls.submind = bot_class(cls.socket, "Private", "testrunner",
                                "testpassword", on_server=False)

    @classmethod
    def tearDownClass(cls) -> None:
        clean_up_bot(cls.submind)
        cls.server.shutdown_server()




import time
from dataclasses import dataclass
from threading import Event
from typing import Dict

from ovos_utils import flatten_list
from ovos_utils.log import LOG
from neon_minerva.intent_services import IntentMatch


EXTENSION_TIME = 15
MIN_RESPONSE_WAIT = 3


@dataclass
class Query:
    session_id: str
    query: str
    replies: list = None
    extensions: list = None
    query_time: float = time.time()
    timeout_time: float = time.time() + 1
    responses_gathered: Event = Event()
    completed: Event = Event()
    answered: bool = False


class CommonQuery:
    def __init__(self, bus):
        self.bus = bus
        self.skill_id = "common_query.test"  # fake skill
        self.active_queries: Dict[str, Query] = dict()
        self._vocabs = {}
        self.bus.on('question:query.response', self.handle_query_response)
        self.bus.on('common_query.question', self.handle_question)
        # TODO: Register available CommonQuery skills

    def is_question_like(self, utterance, lang):
        # skip utterances with less than 3 words
        if len(utterance.split(" ")) < 3:
            return False
        return True

    def match(self, utterances, lang, message):
        """Send common query request and select best response

        Args:
            utterances (list): List of tuples,
                               utterances and normalized version
            lang (str): Language code
            message: Message for session context
        Returns:
            IntentMatch or None
        """
        # we call flatten in case someone is sending the old style list of tuples
        utterances = flatten_list(utterances)
        match = None
        for utterance in utterances:
            if self.is_question_like(utterance, lang):
                message.data["lang"] = lang  # only used for speak
                message.data["utterance"] = utterance
                answered = self.handle_question(message)
                if answered:
                    match = IntentMatch('CommonQuery', None, {}, None,
                                        utterance)
                break
        return match

    def handle_question(self, message):
        """
        Send the phrase to the CommonQuerySkills and prepare for handling
        the replies.
        """
        utt = message.data.get('utterance')
        sid = "test_session"
        # TODO: Why are defaults not creating new objects on init?
        query = Query(session_id=sid, query=utt, replies=[], extensions=[],
                      query_time=time.time(), timeout_time=time.time() + 1,
                      responses_gathered=Event(), completed=Event(),
                      answered=False)
        assert query.responses_gathered.is_set() is False
        assert query.completed.is_set() is False
        self.active_queries[sid] = query

        LOG.info(f'Searching for {utt}')
        # Send the query to anyone listening for them
        msg = message.reply('question:query', data={'phrase': utt})
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg)

        query.timeout_time = time.time() + 1
        timeout = False
        while not query.responses_gathered.wait(EXTENSION_TIME):
            if time.time() > query.timeout_time + 1:
                LOG.debug(f"Timeout gathering responses ({query.session_id})")
                timeout = True
                break

        # forcefully timeout if search is still going
        if timeout:
            LOG.warning(f"Timed out getting responses for: {query.query}")
        self._query_timeout(message)
        if not query.completed.wait(10):
            raise TimeoutError("Timed out processing responses")
        answered = bool(query.answered)
        self.active_queries.pop(sid)
        LOG.debug(f"answered={answered}|"
                  f"remaining active_queries={len(self.active_queries)}")
        return answered

    def handle_query_response(self, message):
        search_phrase = message.data['phrase']
        skill_id = message.data['skill_id']
        searching = message.data.get('searching')
        answer = message.data.get('answer')

        query = self.active_queries.get("test_session")
        if not query:
            LOG.warning(f"No active query for: {search_phrase}")
        # Manage requests for time to complete searches
        if searching:
            LOG.debug(f"{skill_id} is searching")
            # request extending the timeout by EXTENSION_TIME
            query.timeout_time = time.time() + EXTENSION_TIME
            # TODO: Perhaps block multiple extensions?
            if skill_id not in query.extensions:
                query.extensions.append(skill_id)
        else:
            # Search complete, don't wait on this skill any longer
            if answer:
                LOG.info(f'Answer from {skill_id}')
                query.replies.append(message.data)

            # Remove the skill from list of timeout extensions
            if skill_id in query.extensions:
                LOG.debug(f"Done waiting for {skill_id}")
                query.extensions.remove(skill_id)

            time_to_wait = query.query_time + MIN_RESPONSE_WAIT - time.time()
            if time_to_wait > 0:
                LOG.debug(f"Waiting {time_to_wait}s before checking extensions")
                query.responses_gathered.wait(time_to_wait)
            # not waiting for any more skills
            if not query.extensions:
                LOG.debug(f"No more skills to wait for ({query.session_id})")
                query.responses_gathered.set()

    def _query_timeout(self, message):
        query = self.active_queries.get("test_session")
        LOG.info(f'Check responses with {len(query.replies)} replies')
        search_phrase = message.data.get('phrase', "")
        if query.extensions:
            query.extensions = []

        # Look at any replies that arrived before the timeout
        # Find response(s) with the highest confidence
        best = None
        ties = []
        for response in query.replies:
            if not best or response['conf'] > best['conf']:
                best = response
                ties = []
            elif response['conf'] == best['conf']:
                ties.append(response)

        if best:
            # invoke best match
            LOG.info('Handling with: ' + str(best['skill_id']))
            cb = best.get('callback_data') or {}
            self.bus.emit(message.forward('question:action',
                                          data={'skill_id': best['skill_id'],
                                                'phrase': search_phrase,
                                                'callback_data': cb}))
            query.answered = True
        else:
            query.answered = False
        query.completed.set()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2023 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2023: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

PROMPT = "hello!"

RESPONSES = {"Ned": "Hi, I'm Ned. How are you, testrunner?",
             "Eliza": "Hello... I'm glad you could drop by today.",
             "terry": "Hey",
             "Ima": "I am ready to talk",
             "wolfram": "Hello, human",
             "kbot": "hello!",
             "alice": "Hi there!",
             "abstain": ""}

VOTES_BY_USER = {
    '1prompt1': {
        'bot1': [],
        'bot2': ['bot1'],
        'bot3': [],
        'bot4': [],
        'abstain': ['bot2', 'bot3', 'bot4']
    },
    '2prompt2': {
        'bot1': [],
        'bot2': [],
        'bot4': [],
        'abstain': ['bot1', 'bot2', 'bot4']
    },
    '3prompt3': {
        'bot1': [],
        'bot2': ['bot1'],
        'bot3': [],
        'bot4': [],
        'abstain': ['bot2', 'bot3', 'bot4']
    }
}

SELECTIONS = {
    'bot1': [('prompt1', 'bot2'), ('prompt2', 'abstain'), ('prompt3', 'bot2')],
    'bot2': [('prompt1', 'abstain'), ('prompt2', 'abstain'), ('prompt3', 'abstain')],
    'bot3': [('prompt1', 'abstain'), ('prompt3', 'abstain')],
    'bot4': [('prompt1', 'abstain'), ('prompt2', 'abstain'), ('prompt3', 'abstain')]
}

SELECTION_HISTORY = ['bot2', 'bot2', 'bot2']

PARTICIPANT_HISTORY = [(),  # element 0 is expected to be an empty tuple
                       ('bot1', 'bot2', 'bot3', 'bot4'),
                       ('bot1', 'bot2', 'bot3', 'bot4')]

# ContextKeeper constants
RECENT_USERS = [f"user{number}" for number in range(20)]
RECENT_SHOUTS = [
    "Who is the president of Ukraine?",
    "Who is the president of the US?",
    "What is the distance between them?",
    "When was he born?"
]



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Optional
from adapt.engine import IntentDeterminationEngine
from ovos_utils.intents.intent_service_interface import open_intent_envelope
from ovos_utils.log import LOG
from ovos_utils.messagebus import FakeBus, get_message_lang

from neon_minerva.exceptions import IntentNotMatched, ConfidenceTooLow
from neon_minerva.intent_services import IntentMatch


class AdaptContainer:
    def __init__(self, lang: str, bus: FakeBus):
        self.lang = lang.lower()
        self.bus = bus
        self.adapt = IntentDeterminationEngine()
        self.bus.on('register_vocab', self.handle_register_vocab)
        self.bus.on('register_intent', self.handle_register_intent)

    def handle_register_vocab(self, message):
        entity_value = message.data.get('entity_value')
        entity_type = message.data.get('entity_type')
        regex_str = message.data.get('regex')
        alias_of = message.data.get('alias_of')
        lang = get_message_lang(message)
        if lang != self.lang:
            return
        if regex_str:
            self.adapt.register_regex_entity(regex_str)
        else:
            self.adapt.register_entity(entity_value, entity_type,
                                       alias_of=alias_of)

    def handle_register_intent(self, message):
        intent = open_intent_envelope(message)
        self.adapt.register_intent_parser(intent)

    def test_intent(self, utterance: str) -> Optional[IntentMatch]:
        best_intent = None
        try:
            intents = [i for i in self.adapt.determine_intent(
                utterance, 100,
                include_tags=True)]
            if intents:
                best_intent = max(intents,
                                  key=lambda x: x.get('confidence', 0.0))
        except Exception as err:
            LOG.exception(err)

        if not best_intent:
            raise IntentNotMatched(utterance)
        LOG.debug(best_intent)
        skill_id = best_intent['intent_type'].split(":")[0]
        _norm_id = skill_id.replace('.', '_')
        intent_data = {k.replace(_norm_id, '', 1): v for k, v in
                       best_intent.items() if k.startswith(_norm_id) and
                       isinstance(v, str)}
        LOG.debug(intent_data)
        ret = IntentMatch('Adapt', best_intent['intent_type'], intent_data,
                          skill_id, utterance)
        return ret
